@@ -6,21 +6,17 @@ import os
 import subprocess
 import shlex
 import sys
+from configparser import RawConfigParser
 from decimal import Decimal
+from functools import lru_cache
 from pathlib import Path
 from platform import platform
 from tempfile import NamedTemporaryFile
 
 try:
-    from ConfigParser import RawConfigParser
-except ImportError:
-    from configparser import RawConfigParser
-
-try:
-    import thread
     import threading
 except ImportError:
-    thread = None
+    threading = None
 
 import argcomplete
 import colorama
@@ -29,8 +25,8 @@ from halo import Halo
 from spinners.spinners import Spinners
 
 from .ansi import MILCFormatter, ansi_colors, ansi_config, ansi_escape, format_ansi
-from .configuration import Configuration, SubparserWrapper, get_argument_name, handle_store_boolean
 from .attrdict import AttrDict
+from .configuration import Configuration, SubparserWrapper, get_argument_name, get_argument_strings, handle_store_boolean
 
 
 class MILC(object):
@@ -40,10 +36,11 @@ class MILC(object):
         """Initialize the MILC object.
         """
         # Setup a lock for thread safety
-        self._lock = threading.RLock() if thread else None
+        self._lock = threading.RLock() if threading else None
 
         # Define some basic info
         self.acquire_lock()
+        self.prog_name = name
         self.version = version
         self.author = author
         self._config_store_true = []
@@ -55,18 +52,23 @@ class MILC(object):
         self._inside_context_manager = False
         self.ansi = ansi_colors
         self.arg_only = {}
-        self.config_file = None
+        self.config_file = self.find_config_file()
         self.default_arguments = {}
         self.platform = platform()
-        self.prog_name = name
         self.interactive = sys.stdin.isatty()
         self.release_lock()
+        self._deprecated_arguments = {}
+        self._deprecated_commands = {}
 
         # Initialize all the things
         self.initialize_config()
         self.initialize_argparse()
         self.initialize_logging()
 
+    @property
+    def config_dir(self):
+        return self.config_file.parent
+
     @property
     def description(self):
         return self._description
@@ -88,12 +90,12 @@ class MILC(object):
             raise RuntimeError('You can only specify *args or **kwargs, not both!')
 
         args = args or kwargs
-        text = format_ansi(text)
+        text = format_ansi(text % args)
 
         if not self.config.general.color:
             text = ansi_escape.sub('', text)
 
-        print(text % args)
+        print(text)
 
     def run(self, command, capture_output=True, combined_output=False, text=True, **kwargs):
         """Run a command using `subprocess.run`, but using some different defaults.
@@ -166,6 +168,7 @@ class MILC(object):
         self._subparsers = None
         self.argwarn = argcomplete.warn
         self.args = AttrDict()
+        self.args_passed = AttrDict()
         self._arg_parser = argparse.ArgumentParser(**kwargs)
         self.set_defaults = self._arg_parser.set_defaults
         self.release_lock()
@@ -186,12 +189,21 @@ class MILC(object):
 
         return self._arg_parser.print_usage(*args, **kwargs)
 
+    def log_deprecated_warning(self, item_type, name, reason):
+        """Logs a warning with a custom message if a argument or command is deprecated.
+        """
+        self.log.warning("Warning: %s '%s' is deprecated:\n\t%s", item_type, name, reason)
+
     def add_argument(self, *args, **kwargs):
         """Wrapper to add arguments and track whether they were passed on the command line.
         """
         if 'action' in kwargs and kwargs['action'] == 'store_boolean':
             return handle_store_boolean(self, *args, **kwargs)
 
+        arg_name = get_argument_name(self._arg_parser, *args, **kwargs)
+        arg_strings = get_argument_strings(self._arg_parser, *args, **kwargs)
+
+        # Handle tab completion
         completer = kwargs.get('completer')
 
         if 'completer' in kwargs:
@@ -204,9 +216,21 @@ class MILC(object):
         else:
             self._arg_parser.add_argument(*args, **kwargs)
 
+        # Record the default for this argument
         if 'general' not in self.default_arguments:
             self.default_arguments['general'] = {}
-        self.default_arguments['general'][get_argument_name(self, *args, **kwargs)] = kwargs.get('default')
+
+        self.default_arguments['general'][arg_name] = kwargs.get('default')
+
+        # Determine if it was passed on the command line
+        if 'general' not in self.args_passed:
+            self.args_passed['general'] = {}
+
+        self.args_passed['general'][arg_name] = False
+
+        for arg in arg_strings:
+            if arg in sys.argv:
+                self.args_passed['general'][arg_name] = True
 
         self.release_lock()
 
@@ -260,6 +284,7 @@ class MILC(object):
         if self._lock:
             self._lock.release()
 
+    @lru_cache(maxsize=None)
     def find_config_file(self):
         """Locate the config file.
         """
@@ -280,7 +305,16 @@ class MILC(object):
         def argument_function(handler):
             config_name = handler.__name__
             subcommand_name = config_name.replace("_", "-")
-            arg_name = get_argument_name(self, *args, **kwargs)
+            arg_name = get_argument_name(self._arg_parser, *args, **kwargs)
+            arg_strings = get_argument_strings(self._arg_parser, *args, **kwargs)
+
+            if 'deprecated' in kwargs:
+                self._deprecated_arguments[arg_name] = kwargs['deprecated']
+                if kwargs['help']:
+                    kwargs['help'] += f" [Deprecated]: {kwargs['deprecated']}"
+                else:
+                    kwargs['help'] = f"[Deprecated]: {kwargs['deprecated']}"
+                del kwargs['deprecated']
 
             if kwargs.get('arg_only'):
                 if arg_name not in self.arg_only:
@@ -297,6 +331,15 @@ class MILC(object):
                 if self.config[config_name][arg_name] is None:
                     self.config[config_name][arg_name] = kwargs.get('default')
 
+                if config_name not in self.args_passed:
+                    self.args_passed[config_name] = {}
+
+                self.args_passed[config_name][arg_name] = False
+
+                for arg in arg_strings:
+                    if arg in sys.argv:
+                        self.args_passed[config_name][arg_name] = True
+
             if handler is self._entrypoint:
                 self.add_argument(*args, **kwargs)
 
@@ -328,15 +371,15 @@ class MILC(object):
 
         self.release_lock()
 
-    def read_config_file(self, config_file):
+    def read_config_file(self):
         """Read in the configuration file and return Configuration objects for it and the config_source.
         """
         config = Configuration()
         config_source = Configuration()
 
-        if config_file.exists():
+        if self.config_file.exists():
             raw_config = RawConfigParser()
-            raw_config.read(str(config_file))
+            raw_config.read(str(self.config_file))
 
             # Iterate over the config file options and write them into config
             for section in raw_config.sections():
@@ -365,8 +408,7 @@ class MILC(object):
         """Read in the configuration file and store it in self.config.
         """
         self.acquire_lock()
-        self.config_file = self.find_config_file()
-        self.config, self.config_source = self.read_config_file(self.config_file)
+        self.config, self.config_source = self.read_config_file()
         self.release_lock()
 
     def merge_args_into_config(self):
@@ -388,24 +430,14 @@ class MILC(object):
             if argument not in self.arg_only or section not in self.arg_only[argument]:
                 # Determine the arg value and source
                 arg_value = getattr(self.args, argument)
-                passed_on_cmdline = False
-
-                if section in self.subcommands:
-                    default_value = self.subcommands[section].get_default(argument)
-                else:
-                    default_value = self._arg_parser.get_default(argument)
-
-                if argument in self._config_store_true and arg_value:
-                    passed_on_cmdline = True
-                elif argument in self._config_store_false and not arg_value:
-                    passed_on_cmdline = True
-                elif arg_value is not None and (self.config[section][argument] is None or arg_value != default_value):
-                    passed_on_cmdline = True
 
                 # Merge this argument into self.config
-                if passed_on_cmdline:
+                if self.args_passed[section][argument] or (argument in self._config_store_true and arg_value) or (argument in self._config_store_false and not arg_value):
                     self.config[section][argument] = arg_value
                     self.config_source[section][argument] = 'argument'
+                elif self.config[section][argument] is None:
+                    # Capture the default value
+                    self.config[section][argument] = arg_value
 
         self.release_lock()
 
@@ -420,13 +452,12 @@ class MILC(object):
                 if self.config_source[section_name][option_name] == 'config_file' and value is not None:
                     sane_config.set(section_name, option_name, str(value))
 
-        config_dir = self.config_file.parent
-        if not config_dir.exists():
-            config_dir.mkdir(parents=True, exist_ok=True)
+        if not self.config_dir.exists():
+            self.config_dir.mkdir(parents=True, exist_ok=True)
 
         # Write the config file atomically.
         self.acquire_lock()
-        with NamedTemporaryFile(mode='w', dir=str(config_dir), delete=False) as tmpfile:
+        with NamedTemporaryFile(mode='w', dir=str(self.config_dir), delete=False) as tmpfile:
             sane_config.write(tmpfile)
 
         if os.path.getsize(tmpfile.name) > 0:
@@ -442,7 +473,7 @@ class MILC(object):
             self.log.warning('%s.config_file not set, not saving config!', self.__class__.__name__)
             return
 
-        config, config_source = self.read_config_file(self.config_file)
+        config, config_source = self.read_config_file()
 
         if section in config and option in config[section] and config[section][option] is None:
             del config[section][option]
@@ -467,6 +498,25 @@ class MILC(object):
         self._save_config_file(self.config)
         self.log.info('Wrote configuration to %s', shlex.quote(str(self.config_file)))
 
+    def check_deprecated(self):
+        entry_name = self._entrypoint.__name__
+
+        if entry_name in self._deprecated_commands:
+            msg = self._deprecated_commands[entry_name]
+            self.log_deprecated_warning('Entrypoint', entry_name, msg)
+
+        if self._subcommand:
+            name = self._subcommand.__name__.replace("_", "-")
+            if name in self._deprecated_commands:
+                msg = self._deprecated_commands[name]
+                self.log_deprecated_warning('Subcommand', name, msg)
+
+        deprecated_args_passed = [arg.replace('-', '') for arg in sys.argv if arg.replace('-', '') in self._deprecated_arguments]
+
+        for arg in deprecated_args_passed:
+            msg = self._deprecated_arguments[arg]
+            self.log_deprecated_warning('Argument', arg, msg)
+
     def __call__(self):
         """Execute the entrypoint function.
         """
@@ -475,6 +525,8 @@ class MILC(object):
             with self:
                 return self.__call__()
 
+        self.check_deprecated()
+
         if self._subcommand:
             return self._subcommand(self)
         elif self._entrypoint:
@@ -482,8 +534,14 @@ class MILC(object):
 
         raise RuntimeError('No entrypoint provided!')
 
-    def entrypoint(self, description):
+    def entrypoint(self, description, deprecated=None):
         """Decorator that marks the entrypoint used when a subcommand is not supplied.
+        Args:
+            description
+                A one-line description to display in --help
+
+            deprecated
+                Deprecation message. When set the subcommand will marked as deprecated and this message will be displayed in the help output.
         """
         if self._inside_context_manager:
             raise RuntimeError('You must run this before cli()!')
@@ -494,6 +552,12 @@ class MILC(object):
 
         def entrypoint_func(handler):
             self.acquire_lock()
+
+            if deprecated:
+                name = handler.__name__
+                self._deprecated_commands[name] = deprecated
+                self.description += f' [Deprecated]: {deprecated}'
+
             self._entrypoint = handler
             self.release_lock()
 
@@ -501,7 +565,7 @@ class MILC(object):
 
         return entrypoint_func
 
-    def add_subcommand(self, handler, description, hidden=False, **kwargs):
+    def add_subcommand(self, handler, description, hidden=False, deprecated=None, **kwargs):
         """Register a subcommand.
 
         Args:
@@ -514,6 +578,10 @@ class MILC(object):
 
             hidden
                 When True don't display this command in --help
+
+            deprecated
+                Deprecation message. When set the subcommand will be marked as deprecated
+                and this message will be displayed in help output.
         """
         if self._inside_context_manager:
             raise RuntimeError('You must run this before the with statement!')
@@ -523,6 +591,10 @@ class MILC(object):
 
         name = handler.__name__.replace("_", "-")
 
+        if deprecated:
+            self._deprecated_commands[name] = deprecated
+            description += f' [Deprecated]: {deprecated}'
+
         self.acquire_lock()
         if not hidden:
             self._subparsers.metavar = "{%s,%s}" % (self._subparsers.metavar[1:-1], name) if self._subparsers.metavar else "{%s%s}" % (self._subparsers.metavar[1:-1], name)
